@@ -1,4 +1,4 @@
-# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
+# Copyright (c) 2020, 2021  Peter Pentchev <roam@ringlet.net>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -36,7 +36,7 @@ be a dictionary). For the present, the keys may only be strings.
 A special case of a dictionary with a single key "*" means any value for
 a key will be accepted. Otherwise, all keys with names not starting with
 a "?" character are mandatory, and any keys with names starting with
-a "?" character are optional.
+a "?" character are optional (or may be present with null values).
 
 The dictionary values may be any of:
 - a Python type signifying that the value must be an instance thereof
@@ -58,7 +58,8 @@ For example, the following schema:
             "meal": set(("breakfast", "lunch", "brunch")),
             "colors": [{
                 "name": str,
-                "intensity": set(["dark", "light"])
+                "intensity": set(["dark", "light"]),
+                "?weight": int,
             }]
         },
         "possessions": {
@@ -80,7 +81,8 @@ For example, the following schema:
             "meal": "brunch",
             "colors": [
                 {"name": "blue", "intensity": "light"},
-                {"name": "green", "intensity": "dark"}
+                {"name": "green", "intensity": "dark", "weight": None},
+                {"name": "red", "intensity": "light", "weight": 75},
             ]
         },
         "possessions": {
@@ -100,7 +102,7 @@ except ImportError:
     pass
 
 
-VERSION = "0.1.1"
+VERSION = "0.2.0"
 
 FEATURES_STRING = "trivval=" + VERSION
 
@@ -190,22 +192,21 @@ def validate_dict(value, schema, flags):
             validate_single(key, value[key], valtype, flags)
         return
 
-    extra = set(value.keys())
+    processed = set()
     for key, valtype in schema.items():
         if key.startswith("?"):
-            required = False
             key = key[1:]
+            if value.get(key, None) is None:
+                processed.add(key)
+                continue
         else:
-            required = True
-
-        if key not in value:
-            if required:
+            if key not in value:
                 raise ValidationError([key], "missing")
-            continue
-        extra.remove(key)
 
+        processed.add(key)
         validate_single(key, value[key], valtype, flags)
 
+    extra = set(value.keys()) - processed
     if extra and not flags & FLAG_ALLOW_EXTRA:
         raise ValidationError([",".join(sorted(extra))], "extra keys")
 
